; ==================================================================================================
; Title:   D3D_Mesh.inc
; Author:  G. Friedrich
; Version: 1.0.0
; Purpose: DirectX Mesh objects for ObjAsm32.
; Notes:   Version 1.0.0, January 2007
;            - First release.
; ==================================================================================================

MESH_LOAD_DATA struc
  pOwner          Pointer   ?
  pDirect3D       Pointer   ?
  pTextureManager Pointer   ?
MESH_LOAD_DATA ends

Object D3D_Mesh, D3D_MeshID, Streamable
  RedefineMethod    Done
  RedefineMethod    Init,             Pointer, Pointer, Pointer
  RedefineMethod    Load,             Pointer, Pointer        ;-> Stream, -> Parameter(s)
  StaticMethod      LoadFromXFile,    Pointer
  StaticMethod      LoadFromXMLFile,  Pointer
  StaticMethod      Render
  StaticMethod      RenderSurface,    dword
  RedefineMethod    Store,            Pointer                 ;-> Stream
  StaticMethod      UpdateBounding

  Private
  DynamicMethod     RenderByAttr,     dword
  PrivateEnd

  DefineVariable    pDirect3D,        Pointer,      NULL      ;-> Direct3D
  DefineVariable    pTextureManager,  Pointer,      NULL      ;-> D3D_TextureManager
  DefineVariable    pIMesh,           Pointer,      NULL      ;-> ID3DXMesh
  DefineVariable    pCritSect,        Pointer,      NULL      ;-> CRITICAL_SECTION

  ;World transform matrices
  DefineVariable    mRotation,        D3DXMATRIX,   {1.0,0,0,0, 0,1.0,0,0, 0,0,0,1.0, 0,0,0,1.0}
  DefineVariable    mTranslation,     D3DXMATRIX,   {1.0,0,0,0, 0,1.0,0,0, 0,0,0,1.0, 0,0,0,1.0}

  ;Mesh bounding sphere based on VertexBuffer data in world coordinates
  DefineVariable    BoundingSphere,   D3DXVECTOR4,  {0.0, 0.0, 0.0, 0.0}   ;x,y,z, radius

  Embed VertexBuffer, D3D_VertexBuffer
  Embed IndexBuffer,  D3D_IndexBuffer
  Embed MaterialColl, DataCollection
  Embed TextureColl,  DwordCollection

ObjectEnd


; ==================================================================================================

if IMPLEMENT

externdef IID_ISAXXMLReader:GUID
externdef CLSID_SAXXMLReader40:GUID

; 
; Method:    D3D_Mesh.Done
; Purpose:   Finalizes the object.
; Arguments: Arg1: -> TextureInterface.
;            Arg2: -> D3D_Texturemanager.
; Return:    Nothing.

TextureRelease proc pTextureData:Pointer, pTextureManager:Pointer
    OCall pTextureManager::D3D_TextureManager.ReleaseTexture, pTextureData
    ret
TextureRelease endp

Method D3D_Mesh.Done, uses esi
    SetObject esi
    SAFE_RELEASE [esi].pIMesh
    OCall [esi].VertexBuffer::D3D_VertexBuffer.Done
    OCall [esi].IndexBuffer::D3D_IndexBuffer.Done
    OCall [esi].MaterialColl::DataCollection.Done
    .if [esi].pTextureManager != NULL
      OCall [esi].TextureColl::DwordCollection.ForEach, offset TextureRelease, [esi].pTextureManager
    .endif
    OCall [esi].TextureColl::DwordCollection.Done
    ACall esi.Done
    .if [esi].pCritSect != NULL
      invoke DeleteCriticalSection, [esi].pCritSect
    .endif
MethodEnd, NONE

; 
; Method:    D3D_Mesh.Init
; Purpose:   Initializes the object.
; Arguments: Arg1: -> Owner object.
;            Arg2: -> Direct3D object.
;            Arg3: -> Texture manager object.
; Return:    Nothing.

Method D3D_Mesh.Init, uses esi, pOwner:Pointer, pDirect3D:Pointer, pTextureManager:Pointer
    SetObject esi
    ACall esi.Init, pOwner
    m2m [esi].pDirect3D, pDirect3D
    m2m [esi].pTextureManager, pTextureManager
    OCall [esi].MaterialColl::DataCollection.Init, esi, 10, 10, COLLECTIONMAXSIZE
    OCall [esi].TextureColl::DwordCollection.Init, esi, 10, 10, COLLECTIONMAXSIZE
MethodEnd

; 
; Method:    D3D_Mesh.Load
; Purpose:   Loads the object from a stream.
; Arguments: Arg1: -> Stream object.
;            Arg2: -> BUFFER_LOAD_DATA.
; Return:    Nothing.

Method D3D_Mesh.Load, uses ebx esi, pStream:Pointer, pLoadData:Pointer
    local dMaxSize:dword, dDelta:dword, dLimit:dword, dCount:dword      ;Reverse order!
    local BufferLoadData:BUFFER_LOAD_DATA
    local dFVF:dword, dVertexCount:dword, dFaceCount:dword, pAttrBuffer:Pointer

    SetObject esi
    mov eax, pLoadData
    m2m [esi].pDirect3D, [eax].MESH_LOAD_DATA.pDirect3D, ecx
    m2m [esi].pTextureManager, [eax].MESH_LOAD_DATA.pTextureManager, ecx
    ACall esi.Init, [eax].MESH_LOAD_DATA.pOwner
    OCall pStream::Stream.BinRead, addr dFaceCount, 3*4

    mov ecx, pLoadData
    lea ebx, BufferLoadData
    mov [ebx].BUFFER_LOAD_DATA.pOwner, esi
    m2m [ebx].BUFFER_LOAD_DATA.pDirect3D, [ecx].MESH_LOAD_DATA.pDirect3D, eax
    mov edx, [ecx].MESH_LOAD_DATA.pDirect3D
    invoke D3DXCreateMeshFVF, dFaceCount, dVertexCount, D3DXMESH_WRITEONLY, \
                              dFVF, [edx].Direct3D.pID3DDevice, addr [esi].pIMesh

    ICall [esi].pIMesh::ID3DXMesh.LockAttributeBuffer, D3DLOCK_READONLY, addr pAttrBuffer
    mov eax, dFaceCount
    shl eax, 2
    OCall pStream::Stream.BinRead, pAttrBuffer, eax
    ICall [esi].pIMesh::ID3DXMesh.UnlockAttributeBuffer
    ICall [esi].pIMesh::ID3DXMesh.GetVertexBuffer, addr [ebx].BUFFER_LOAD_DATA.pIBuffer
    OCall [esi].VertexBuffer::D3D_VertexBuffer.Load, pStream, ebx
    ICall [esi].pIMesh::ID3DXMesh.GetIndexBuffer, addr [ebx].BUFFER_LOAD_DATA.pIBuffer
    OCall [esi].IndexBuffer::D3D_IndexBuffer.Load, pStream, ebx
    OCall [esi].MaterialColl::DataCollection.Load, pStream, esi

    OCall pStream::Stream.BinRead, addr dCount, 4 * 4    ;Count, ..., MaxSize
    OCall [esi].TextureColl::DwordCollection.Init, esi, dLimit, dDelta, dMaxSize
    .while dCount > 0
      OCall pStream::Stream.StrRead
      push eax
      .if [esi].pTextureManager != NULL
        OCall [esi].pTextureManager::D3D_TextureManager.GetTexture, eax
      .else
        xor eax, eax
      .endif
      OCall [esi].TextureColl::DwordCollection.Insert, eax
      call StrDispose
      dec dCount
    .endw

    ;Set buffer markers
    mov [esi].VertexBuffer.dFirstPos, 0
    m2m [esi].VertexBuffer.dLastPos, [esi].VertexBuffer.dCapacity
    mov [esi].IndexBuffer.dFirstPos, 0
    m2m [esi].IndexBuffer.dLastPos, [esi].IndexBuffer.dCapacity
MethodEnd

; 
; Method:    D3D_Mesh.LoadFromXFile
; Purpose:   Loads a mesh from a .x file.
; Arguments: Arg1: -> UNICODE FileName.
; Return:    HRESULT.

Method D3D_Mesh.LoadFromXFile, uses ebx esi, pFileName:Pointer
    local pIMaterialBuffer:Pointer
    local pIAdjacency:Pointer, pMaterials:Pointer, dMaterialCount:dword, pMaterial:Pointer
    local pIBuffer:Pointer, pBuffer:Pointer
    local bTexPath[1024]:word, pNamePos:Pointer

    SetObject esi
    lea ebx, bTexPath
    invoke UStr2Str, ebx, pFileName
    invoke GetFullPathName, ebx, sizeof bTexPath - 1, addr bTexPath, addr pNamePos
    mov eax, pNamePos
    .if eax == NULL
      mov pNamePos, ebx
    .endif

    .if [esi].pIMesh != NULL
      ICall [esi].pIMesh::IUnknown.Release
      mov [esi].pIMesh, 0
    .endif
    OCall [esi].VertexBuffer::D3D_VertexBuffer.Done
    OCall [esi].IndexBuffer::D3D_IndexBuffer.Done
    mov ecx, [esi].pDirect3D
    invoke D3DXLoadMeshFromXW, pFileName, 0, [ecx].Direct3D.pID3DDevice, NULL, \
              addr pIMaterialBuffer, addr pIAdjacency, addr dMaterialCount, addr [esi].pIMesh

    .if SUCCEEDED(eax)
      OCall [esi].TextureColl::DwordCollection.ForEach, TextureRelease, [esi].pTextureManager
      OCall [esi].TextureColl::DwordCollection.DeleteAll

      mov pMaterials, $ICall(pIMaterialBuffer::ID3DXBuffer.GetBufferPointer)

      ;Fill the material DataCollection
      OCall [esi].MaterialColl::DataCollection.DisposeAll
      xor ebx, ebx
      .while ebx < dMaterialCount
        MemAlloc sizeof D3DMATERIAL9
        .if eax != NULL
          mov pMaterial, eax
          .if [esi].pTextureManager != NULL
            mov eax, pMaterials
            invoke StrCopy, pNamePos, [eax].D3DXMATERIAL.pTextureFilename
            OCall [esi].pTextureManager::D3D_TextureManager.GetTexture, addr bTexPath
            OCall [esi].TextureColl::DwordCollection.Insert, eax
          .endif
          invoke MemClone, pMaterial, pMaterials, sizeof D3DMATERIAL9
          ;Set the ambient color for the material (D3DX does not do this)
          mov eax, pMaterial
          m2m [eax].D3DMATERIAL9.Ambient.r, [eax].D3DMATERIAL9.Diffuse.r, ecx
          m2m [eax].D3DMATERIAL9.Ambient.g, [eax].D3DMATERIAL9.Diffuse.g, ecx
          m2m [eax].D3DMATERIAL9.Ambient.b, [eax].D3DMATERIAL9.Diffuse.b, ecx
          m2m [eax].D3DMATERIAL9.Ambient.a, [eax].D3DMATERIAL9.Diffuse.a, ecx
          OCall [esi].MaterialColl::DataCollection.Insert, eax
          add pMaterials, sizeof D3DXMATERIAL
          inc ebx
        .else
          OCall esi.ErrorReport, NULL, D3D_OUT_OF_MEMORY
          .break
        .endif
      .endw
      ICall pIMaterialBuffer::ID3DXBuffer.Release

      ;Update Vertex Buffer
      ICall [esi].pIMesh::ID3DXMesh.GetVertexBuffer, addr pIBuffer
      OCall [esi].VertexBuffer::D3D_VertexBuffer.Init, esi, [esi].pDirect3D, pIBuffer
      ICall pIBuffer::IDirect3DVertexBuffer9._Lock, 0, 0, addr pBuffer, D3DLOCK_READONLY
      xor edx, edx
      mov eax, [esi].VertexBuffer.dVertexSize
      mov ecx, [esi].VertexBuffer.dCapacity
      mul ecx
      invoke MemClone, [esi].VertexBuffer.pBuffer, pBuffer, eax
      ICall pIBuffer::IDirect3DVertexBuffer9.Unlock

      ;Update Index Buffer
      ICall [esi].pIMesh::ID3DXMesh.GetIndexBuffer, addr pIBuffer
      OCall [esi].IndexBuffer::D3D_IndexBuffer.Init, esi, [esi].pDirect3D, pIBuffer
      ICall pIBuffer::IDirect3DIndexBuffer9._Lock, 0, 0, addr pBuffer, D3DLOCK_READONLY
      xor edx, edx
      mov eax, [esi].IndexBuffer.dIndexSize
      mov ecx, [esi].IndexBuffer.dCapacity
      mul ecx
      invoke MemClone, [esi].IndexBuffer.pBuffer, pBuffer, eax
      ICall pIBuffer::IDirect3DIndexBuffer9.Unlock
    .else
      OCall esi.ErrorReport, NULL, D3D_FILE_LOAD_FAILED
    .endif
MethodEnd, eax

; 
; Method:    D3D_Mesh.LoadFromXMLFile
; Purpose:   Loads a mesh from a .3dxml file.
; Arguments: Arg1: -> FileName.
; Return:    HRESULT.

AddCount proc pDwordCollection:Pointer, pSum:Pointer
    mov eax, pDwordCollection
    mov ecx, [eax].DwordCollection.dCount
    mov eax, pSum
    add [eax], ecx
    ret
AddCount endp

Method D3D_Mesh.LoadFromXMLFile, uses ebx edi esi, pFileName:Pointer
    local ContentHandler:XML3DContentHandler
    local PositionStorageColl:Collection, NormalStorageColl:Collection, IndexStorageColl:Collection
    local AttrColl:DwordCollection
    local pXMLReader:Pointer, dSum:dword, dFaceCount:dword, dVertexCount:dword
    local pAttrData:Pointer, pBuffer:Pointer, dCount:dword, wIndexOffset:word, dCurrAttr:dword

    SetObject esi
    invoke CoCreateInstance, offset CLSID_SAXXMLReader40, NULL, \   ;Create an instance of a
              CLSCTX_INPROC_SERVER, offset IID_ISAXXMLReader, \     ;SAXXMLReader40 object
              addr pXMLReader
    .if SUCCEEDED(eax)                                              ;If creation was successful
      SAFE_RELEASE [esi].pIMesh

      ;Initialize the recieving containers
      LNew PositionStorageColl, Collection, Init, esi, 10, 10, COLLECTIONMAXSIZE
      LNew NormalStorageColl, Collection, Init, esi, 10, 10, COLLECTIONMAXSIZE
      LNew IndexStorageColl, Collection, Init, esi, 10, 10, COLLECTIONMAXSIZE
      LNew AttrColl, DwordCollection, Init, esi, 100, 100, COLLECTIONMAXSIZE
      OCall [esi].MaterialColl::DataCollection.DisposeAll
      LNew ContentHandler, XML3DContentHandler, Init, esi, [esi].pDirect3D, \
                                  addr PositionStorageColl, addr NormalStorageColl, \
                                  addr IndexStorageColl, addr [esi].MaterialColl

      mov AttrColl.DwordCollection.dDuplicates, TRUE
      ;Invoke the XML reader
      ICall pXMLReader::ISAXXMLReader.putContentHandler, addr ContentHandler
      ICall pXMLReader::ISAXXMLReader.parseURL, pFileName
      push eax
      ICall pXMLReader::ISAXXMLReader.Release
      pop eax
      .if SUCCEEDED(eax)
        ;Process the recieved data
        mov dSum, 0
        OCall IndexStorageColl::Collection.ForEach, addr AddCount, addr dSum
        mov eax, dSum
        xor edx, edx
        mov ecx, 3
        div ecx
        mov dFaceCount, eax

        mov dSum, 0
        OCall PositionStorageColl::Collection.ForEach, addr AddCount, addr dSum
        mov eax, dSum
        xor edx, edx
        mov ecx, 3
        div ecx
        mov dVertexCount, eax

        mov ecx, [esi].pDirect3D
        invoke D3DXCreateMeshFVF, dFaceCount, dVertexCount, \
                                  D3DXMESH_WRITEONLY, D3DFVF_XYZ or D3DFVF_NORMAL, \
                                  [ecx].Direct3D.pID3DDevice, addr [esi].pIMesh
        .if SUCCEEDED(eax)
          ICall [esi].pIMesh::ID3DXMesh.GetVertexBuffer, addr pBuffer
          OCall [esi].VertexBuffer::D3D_VertexBuffer.Init, esi, [esi].pDirect3D, pBuffer

          ;Load the VertexBuffer with the obtained data
          mov edi, [esi].VertexBuffer.pBuffer
          xor ebx, ebx
          push esi
          .while ebx < PositionStorageColl.Collection.dCount
            push ebx
            OCall PositionStorageColl::Collection.ItemAt, ebx
            mov esi, [eax].DwordCollection.pItems
            mov eax, [eax].DwordCollection.dCount
            xor edx, edx
            mov ecx, 3
            div ecx
            mov dCount, eax
            OCall NormalStorageColl::Collection.ItemAt, ebx
            mov edx, [eax].DwordCollection.pItems

            ;Shuffle position and normal vectors
            xor ebx, ebx
            .while ebx < dCount
              mov ecx, 3
              rep movsd
              xchg edx, esi
              mov ecx, 3
              rep movsd
              xchg edx, esi
              inc ebx
            .endw
            pop ebx
            inc ebx
          .endw
          pop esi
          mov [esi].VertexBuffer.dFirstPos, 0
          m2m [esi].VertexBuffer.dLastPos, [esi].VertexBuffer.dCapacity

          ICall [esi].pIMesh::ID3DXMesh.GetIndexBuffer, addr pBuffer
          OCall [esi].IndexBuffer::D3D_IndexBuffer.Init, esi, [esi].pDirect3D, pBuffer

          ;Load the IndexBuffer with the obtained data
          mov edi, [esi].IndexBuffer.pBuffer
          xor ebx, ebx
          mov wIndexOffset, bx
          push esi
          .while ebx < IndexStorageColl.Collection.dCount
            mov dCurrAttr, ebx
            push ebx
            OCall IndexStorageColl::Collection.ItemAt, ebx
            m2m dCount, [eax].DwordCollection.dCount
            mov esi, [eax].DwordCollection.pItems

            ;D3DFMT_INDEX16 => Scale down from dword to word
            xor ebx, ebx
            xor eax, eax
            .while ebx < dCount
              mov edx, dword ptr [esi]
              add dx, wIndexOffset
              mov [edi], dx
              add esi, 4
              add edi, 2
              inc ebx
              .if eax == 2
                OCall AttrColl::DwordCollection.Insert, dCurrAttr
                xor eax, eax
              .else
                inc eax
              .endif
            .endw
            pop ebx

            OCall PositionStorageColl::Collection.ItemAt, ebx
            mov eax, [eax].DwordCollection.dCount
            xor edx, edx
            mov ecx, 3
            div ecx
            add wIndexOffset, ax

            inc ebx
          .endw
          pop esi
          mov [esi].IndexBuffer.dFirstPos, 0
          m2m [esi].IndexBuffer.dLastPos, [esi].IndexBuffer.dCapacity

          ;Update Attribute buffer
          ICall [esi].pIMesh::ID3DXMesh.LockAttributeBuffer, D3DLOCK_DISCARD, addr pAttrData
          mov eax, AttrColl.DwordCollection.dCount
          shl eax, 2
          invoke MemClone, pAttrData, AttrColl.DwordCollection.pItems, eax
          ICall [esi].pIMesh::ID3DXMesh.UnlockAttributeBuffer

          ;Load Texture Collection with NULLs
          xor ebx, ebx
          .while ebx < [esi].MaterialColl.dCount
            OCall [esi].TextureColl::DwordCollection.Insert, NULL
            inc ebx
          .endw

        .else
          OCall esi.ErrorReport, NULL, D3D_MESH_CREATE_FAILED
        .endif
      .else
        OCall esi.ErrorReport, NULL, D3D_INVALID_3DXML_FILE
      .endif
      ;Housekeeping
      OCall PositionStorageColl::Collection.Done
      OCall NormalStorageColl::Collection.Done
      OCall IndexStorageColl::Collection.Done
    .else
      OCall esi.ErrorReport, NULL, D3D_FILE_LOAD_FAILED
    .endif
MethodEnd

; 
; Method:    D3D_Mesh.Render
; Purpose:   Renders the mesh.
; Arguments: Nothing.
; Return:    HRESULT.

Method D3D_Mesh.Render, uses ebx esi
    SetObject esi
    .if [esi].pIMesh != NULL && [esi].dErrorCode == OBJ_OK
      OCall [esi].pDirect3D::Direct3D.SetRS, D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL
      OCall [esi].VertexBuffer::D3D_VertexBuffer.Update
      OCall [esi].IndexBuffer::D3D_IndexBuffer.Update
      mov eax, [esi].pDirect3D
      ICall [eax].Direct3D.pID3DDevice::IDirect3DDevice9.SetFVF, [esi].VertexBuffer.dFVF
      xor ebx, ebx
      .while ebx < [esi].MaterialColl.dCount
        OCall esi.RenderByAttr, ebx
        inc ebx
      .endw
    .endif
    xor eax, eax                ;Return S_OK
MethodEnd, eax

; 
; Method:    D3D_Mesh.RenderByAttr
; Purpose:   (Private) Renders a mesh surface given an attribute ID.
; Arguments: Arg1: Attribute ID.
; Return:    Nothing.

Method D3D_Mesh.RenderByAttr, uses esi, dAttrID:dword
    SetObject esi
    OCall [esi].MaterialColl::DataCollection.ItemAt, dAttrID
    .if eax != NULL
      mov ecx, [esi].pDirect3D
      ICall [ecx].Direct3D.pID3DDevice::IDirect3DDevice9.SetMaterial, eax
    .endif
    mov eax, dAttrID
    .if eax < [esi].TextureColl.dCount
      OCall [esi].TextureColl::DwordCollection.ItemAt, dAttrID
      mov ecx, [esi].pDirect3D
      ICall [ecx].Direct3D.pID3DDevice::IDirect3DDevice9.SetTexture, 0, [eax].TEXTURE_DATA.pInterface
    .endif
    ICall [esi].pIMesh::ID3DXMesh.DrawSubset, dAttrID
MethodEnd

; 
; Method:    D3D_Mesh.RenderSurface
; Purpose:   Renders the mesh.
; Arguments: Arg1: Surface ID.
; Return:    HRESULT.

Method D3D_Mesh.RenderSurface, uses esi, dID:dword
    SetObject esi
    .if [esi].pIMesh != NULL && [esi].dErrorCode == OBJ_OK
      OCall [esi].pDirect3D::Direct3D.SetRS, D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL
      OCall [esi].VertexBuffer::D3D_VertexBuffer.Update
      OCall [esi].IndexBuffer::D3D_IndexBuffer.Update
      mov eax, [esi].pDirect3D
      ICall [eax].Direct3D.pID3DDevice::IDirect3DDevice9.SetFVF, [esi].VertexBuffer.dFVF
      mov eax, dID
      .if eax < [esi].MaterialColl.dCount
        OCall esi.RenderByAttr, dID
      .else
        OCall esi.ErrorReport, NULL, D3D_INVALID_INDEX
      .endif
    .endif
    xor eax, eax                ;Return S_OK
MethodEnd, eax

; 
; Method:    D3D_Mesh.Store
; Purpose:   Saves the object in a stream.
; Arguments: Arg1: -> Stream
; Return:    Nothing.

StoreTexture proc pTextureData:Pointer, pStream:Pointer
    mov eax, pTextureData
    OCall pStream::Stream.StrWrite, [eax].TEXTURE_DATA.pSourcePath
    ret
StoreTexture endp

Method D3D_Mesh.Store, uses esi, pStream:Pointer
    local dFaceCount:dword, pAttrBuffer:Pointer

    SetObject esi
    ;Store the required parameters to create the mesh
    xor edx, edx
    mov eax, [esi].IndexBuffer.dCapacity
    mov ecx, 3
    div ecx
    mov dFaceCount, eax
    OCall pStream::Stream.BinWrite, addr dFaceCount, 4
    OCall pStream::Stream.BinWrite, addr [esi].VertexBuffer.dCapacity, 4
    OCall pStream::Stream.BinWrite, addr [esi].VertexBuffer.dFVF, 4

    ;Store mesh buffers and related information
    ICall [esi].pIMesh::ID3DXMesh.LockAttributeBuffer, D3DLOCK_READONLY, addr pAttrBuffer
    mov eax, dFaceCount
    shl eax, 2
    OCall pStream::Stream.BinWrite, pAttrBuffer, eax
    ICall [esi].pIMesh::ID3DXMesh.UnlockAttributeBuffer
    OCall [esi].VertexBuffer::D3D_VertexBuffer.Store, pStream
    OCall [esi].IndexBuffer::D3D_IndexBuffer.Store, pStream
    OCall [esi].MaterialColl::DataCollection.Store, pStream
    OCall pStream::Stream.BinWrite, addr [esi].TextureColl.dCount, 4 * 4    ;Count, ..., MaxSize
    OCall [esi].TextureColl::DwordCollection.ForEach, offset StoreTexture, pStream
MethodEnd

; 
; Method:    D3D_Mesh.UpdateBounding
; Purpose:   Updates the internal data of the bounding sphere.
; Arguments: None.
; Return:    Nothing.
; Note:      The bounding sphere is in world space.

Method D3D_Mesh.UpdateBounding, uses esi
    SetObject esi
    invoke D3DXComputeBoundingSphere, [esi].VertexBuffer.pBuffer, [esi].VertexBuffer.dCapacity, \
           [esi].VertexBuffer.dVertexSize, addr [esi].BoundingSphere, addr [esi].BoundingSphere.w
    lea ecx, [esi].BoundingSphere
    invoke D3DXVec3TransformCoord, ecx, ecx, addr [esi].mTranslation
MethodEnd; 4, NONE

endif